/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: ev4.c,v $
 *
 * HISTORY
 * $Log: ev4.c,v $
 * Revision 1.6  1996/01/30  22:54:40  slm
 * Updated copyright and version.
 *
 * Revision 1.5  1995/03/10  19:34:19  davidm
 * Request user-input before dumping performance counters.
 *
 * Revision 1.4  1994/11/20  01:39:09  davidm
 * Performance monitor code now excluded unless SCOUT_PERFMON is defined.
 *
 * Revision 1.3  1994/07/04  03:18:50  davidm
 * Various fixes.  Added support for performance counter based profiling.
 *
 * Revision 1.2  1994/06/28  19:11:58  davidm
 * Minor fix and enabled PERF_CNT_H[0] and PERF_CNT_H[1].  The former
 * counts B-cache misses on the Flamingo/Sandpiper, the latter B-cache
 * misses that require victim writes.
 *
 * Revision 1.1  1994/06/28  18:43:44  davidm
 * Initial revision
 */
/*
 * EV4 (21064) support for performance counters.
 *
 * This is based on the following documentation:
 *
 * DECchip 21064-AA Microprocessor Hardware Reference Manual, DEC
 * Order #: EC-N0079-72 (also available online at gatekeeper.dec.com).
 */
#include <strings.h>
#include "scout_sync_arch.h"
#include "alpha_pal.h"
#include "ev4.h"

#ifdef SCOUT_PERFMON

/*
 * These definitions are CPU implementation specific (i.e., not
 * architecturally defined).
 */
#define PAL_PERFMON_DISABLE		0
#define PAL_PERFMON_ENABLE		1
#define PAL_PERFMON_SET_MUX		2
#define PAL_PERFMON_SET_OPTIONS		3

#define PAL_PERFMON_OPT_ALL_PROC	(1<<0)	/* monitor all processes */
#define PAL_PERFMON_OPT_PME_PROC	(1<<1)	/* only proc with PME set */

extern u_long ev4_current_time;	/* 64 bit current time from alpha_forward.s */
extern u_int  ev4_last_time;	/* last time above was updated (32 bit cc) */
extern int    ev4_selected_counter;

static int	event_source = 0;
static int	event_source_mask = (EV4_ISSUES | EV4_DRY     | EV4_LOADS |
				     EV4_FROZEN | EV4_BRANCH  | EV4_PALMODE |
				     EV4_NONISS | EV4_CNT0    | EV4_DCACHE |
				     EV4_ICACHE | EV4_DUALISS | EV4_MISPR |
				     EV4_INTOPS | EV4_STORES  | EV4_CNT1);
/*
 * The only meaningful performance data we can obtain with the EV4
 * counters are event *rates*.  We measure those by counting the
 * number of cycles that expire during two consecutive counter
 * overflows.  This is complicated by the fact that there are
 * intervals during which we don't want the counters to advance
 * (e.g., when the idle-thread is running).  So those intervals
 * should not be included in the interval length.  To achieve
 * this, we have time_sofar to keep track of how much time
 * has passed during earlier periods during which the counters
 * were enabled.  The difference (read_itimer() - start_time) gives
 * the time spent in the most recent interval (assuming the counters
 * are enabled).
 */
static u_long	start_time;		/* starting time */
static u_long	time_sofar = 0;		/* total time so far */
static int	counters_enabled = 0;	/* just for a sanity check */

static struct event_source {
    char	*name;
    int		nevents_per_intr;
    u_long	nevents;
    u_long	time;
} cnt[NUM_EVENT_SOURCES] = {
    {"Total Issues/2:        ", 16*4096},	/* event-source 0 */
    {"Pipeline Dry:          ", 16*4096},	/* event-source 1 */
    {"Load Instructions:     ", 16*4096},	/* event-source 2 */
    {"Pipeline Frozen:       ", 16*4096},	/* event-source 3 */
    {"Branch Instructions:   ", 16*4096},	/* event-source 4 */
    {"PALmode cycles:        ", 16*4096},	/* event-source 5 */
    {"Total Non-issues/2:    ", 16*4096},	/* event-source 6 */
    {"PERF_CNT_H[0]:         ", 16*4096},	/* event-source 7 */
    {"D-cache misses:        ", 16*4096},	/* event-source 8 */
    {"I-cache misses:        ", 16*4096},	/* event-source 9 */
    {"Dual issues:           ", 16*4096},	/* event-source 10 */
    {"Branch Mispredicts:    ", 16*4096},	/* event-source 11 */
    {"FP Op Instructions:    ", 16*4096},	/* event-source 12 */
    {"Integer Op Instr.:     ", 16*4096},	/* event-source 13 */
    {"Store Instructions:    ", 16*4096},	/* event-source 14 */
    {"PERF_CNT_H[1]:         ", 16*4096}	/* event-source 15 */
};


/*
 * Write the performance monitor register.
 */
static void
wrperfmon(long op, long value)
{
    register u_long a0 asm ("16"), a1 asm ("17"), v0 asm ("0");

    a0 = (op); a1 = (value);

    asm volatile ("call_pal %1 # %0 %2 %3"
		  : "r="(v0)
		  : "i"(PAL_wrperfmon), "r"(a0), "r"(a1)
		  : "memory", "1", "16", "17", "22", "23", "24", "25");
    if (v0 != 1 PREDICT_FALSE) {
	printf("wrperfmon: unexpected return value %ld\n", v0);
    } /* if */
} /* wrperfmon */


static void
select_event_source(long event_source)
{
    u_long mask;

    /*
     * When selecting an event source on counter x, we select a
     * slow event source on counter y and vice versa in order
     * to minimize the interrupt load low.
     */
    if (event_source < 8) {
	mask = 0x108L | (event_source<<9) | (                7L<<32);
	ev4_selected_counter = 0;
    } else {
	mask = 0x000L | (           7<<9) | ((event_source - 8)<<32);
	ev4_selected_counter = 1;
    } /* if */
    wrperfmon(PAL_PERFMON_SET_MUX, mask);
} /* select_event_source */


static void
next_event_source(void)
{
    if (!event_source_mask PREDICT_FALSE) {
	scout_panic("next_event_source: no active event-source");
    } /* if */

    if ((event_source_mask & (event_source_mask - 1)) != 0) {
	/* more than one event-source active: search for next one: */
	do {
	    event_source = (event_source + 1) & (NUM_EVENT_SOURCES - 1);
	} while (!(event_source_mask & (1<<event_source)));
    } /* if */
    select_event_source(event_source);
} /* next_event_source */


void
ev4_perfmon_interrupt(void)
{
    static int previous_event_source = -1;
    struct event_source *esp;

    if (!counters_enabled PREDICT_FALSE) {
	printf("ev4_perfmon_intr: got interrupt while counters disabled!\n");
	return;
    } /* if */

    time_sofar += (ev4_current_time - start_time);
    esp = &cnt[event_source];
    if (previous_event_source != event_source) {
	previous_event_source = event_source;
    } else {
	if (time_sofar > esp->nevents_per_intr PREDICT_TRUE) {
	    esp->time	 += time_sofar;
	    esp->nevents += esp->nevents_per_intr;
	} else {
	    printf("ev4_perfmon_intr: time=%lu is too small (%s%u)\n",
		   time_sofar, esp->name, esp->nevents_per_intr);
	} /* if */
	next_event_source();
    } /* if */
    time_sofar = 0;
    start_time = ev4_current_time;
} /* ev4_perfmon_interrupt */


void
ev4_enable_counters(void)
{
    if (!counters_enabled) {
	counters_enabled = 1;
	start_time = ev4_current_time;
	ev4_last_time = read_itimer();
	wrperfmon(PAL_PERFMON_SET_OPTIONS, PAL_PERFMON_OPT_ALL_PROC);
    } /* if */
} /* ev4_enable_counters */


void
ev4_disable_counters(void)
{
    Scout_IPL old_ipl;
    u_int now;

    if (counters_enabled) {
	wrperfmon(PAL_PERFMON_SET_OPTIONS, 0);
	now = read_itimer();
	/*
	 * Even tough the counters are stopped, there still might
	 * be an interrupt that can hit us, so better make sure this
	 * won't confuse us:
	 */
	old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);

	time_sofar += (now - ev4_last_time) + (ev4_current_time - start_time);
	counters_enabled = 0;

	scout_swap_ipl(old_ipl);
    } /* if */
} /* ev4_disable_counters */


/*
 * Enable/disable tracking of counter COUNTER.  If COUNTER < 0,
 * all performance monitoring is either enabled or disabled.
 */
void
ev4_perfmon_ctl (int mask)
{
    if (mask == EV4_PERFMON_ON || mask == EV4_PERFMON_OFF) {
	Scout_IPL old_ipl;
	/*
	 * NOTICE: we *have* to do a swpipl PAL call to have the
	 * changes take effect!
	 */
	old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);
	wrperfmon(PAL_PERFMON_DISABLE, 0);
	if (mask == EV4_PERFMON_ON) {
	    event_source = ffs(event_source_mask) - 1;
	    select_event_source(event_source);
	    ev4_enable_counters();
	    wrperfmon(PAL_PERFMON_ENABLE, 0);
	} /* if */
	scout_swap_ipl(old_ipl);
    } else {
	event_source_mask = mask;
    } /* if */
} /* ev4_perfmon_ctl */


void
ev4_perfmon_dump (void)
{
    int i;
    void *sl;
    char line[128];

    /* make sure sampling is off: */
    ev4_perfmon_ctl(EV4_PERFMON_OFF);

    printf("ev4: press return when ready to dump performance counters: ");
    gets(line);

    printf("ev4: dumping performance counters\n");

    sl = slog_open("perfmon.out", "w");
    strcpy(line, "# name\t\t\tnevents\t\tcycles\n");
    if (sl) {
	slog_write(sl, line, strlen(line));
    } else {
	printf(line);
    } /* if */
    for (i = 0; i < NUM_EVENT_SOURCES; ++i) {
	if (event_source_mask & (1<<i)) {
	    sprintf(line, "%s\t%lu\t%lu\n",
		    cnt[i].name, cnt[i].nevents, cnt[i].time);
	    if (sl) {
		slog_write(sl, line, strlen(line));
	    } else {
		printf(line);
	    } /* if */
	    cnt[i].nevents = 0;
	    cnt[i].time = 0;
	} /* if */
    } /* for */
    slog_close(sl);

    printf("ev4: press return when ready to continue testing: ");
    gets(line);
} /* ev4_perfmon_dump */

#endif /* SCOUT_PERFMON */

			/*** end of ev4.c ***/
